@@ -1,9 +1,9 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
+
 from dateutil.relativedelta import relativedelta
-from trytond.model import ModelView, ModelSQL, fields
-from trytond.wizard import Wizard, StateView, StateTransition, StateAction, \
-    Button
+from trytond.model import ModelView, ModelSQL, Workflow, fields
+from trytond.wizard import Wizard, StateView, StateAction, Button
 from trytond.tools import datetime_strftime
 from trytond.pyson import Eval, If, PYSONEncoder
 from trytond.transaction import Transaction
@@ -11,29 +11,31 @@ from trytond.pool import Pool
 
 __all__ = ['FiscalYear',
     'BalanceNonDeferralStart', 'BalanceNonDeferral',
-    'CloseFiscalYearStart', 'CloseFiscalYear']
+    'RenewFiscalYearStart', 'RenewFiscalYear']
 
 STATES = {
-    'readonly': Eval('state') == 'close',
+    'readonly': Eval('state') != 'open',
 }
 DEPENDS = ['state']
 
 
-class FiscalYear(ModelSQL, ModelView):
+class FiscalYear(Workflow, ModelSQL, ModelView):
     'Fiscal Year'
     __name__ = 'account.fiscalyear'
     name = fields.Char('Name', size=None, required=True, depends=DEPENDS)
-    code = fields.Char('Code', size=None)
     start_date = fields.Date('Starting Date', required=True, states=STATES,
-            depends=DEPENDS)
+        domain=[('start_date', '<=', Eval('end_date', None))],
+        depends=DEPENDS + ['end_date'])
     end_date = fields.Date('Ending Date', required=True, states=STATES,
-            depends=DEPENDS)
+        domain=[('end_date', '>=', Eval('start_date', None))],
+        depends=DEPENDS + ['start_date'])
     periods = fields.One2Many('account.period', 'fiscalyear', 'Periods',
             states=STATES, depends=DEPENDS)
     state = fields.Selection([
-        ('open', 'Open'),
-        ('close', 'Close'),
-        ], 'State', readonly=True, required=True)
+            ('open', 'Open'),
+            ('close', 'Close'),
+            ('locked', 'Locked'),
+            ], 'State', readonly=True, required=True)
     post_move_sequence = fields.Many2One('ir.sequence', 'Post Move Sequence',
             required=True, domain=[('code', '=', 'account.move'),
                 ['OR',
@@ -50,6 +52,7 @@ class FiscalYear(ModelSQL, ModelView):
             ('id', If(Eval('context', {}).contains('company'), '=', '!='),
                 Eval('context', {}).get('company', -1)),
             ], select=True)
+    icon = fields.Function(fields.Char("Icon"), 'get_icon')
 
     @classmethod
     def __setup__(cls):
@@ -70,20 +73,33 @@ class FiscalYear(ModelSQL, ModelView):
                 'reopen_error': ('You can not reopen fiscal year "%s" until '
                     'you reopen all later fiscal years.'),
                 })
+        cls._transitions |= set((
+                ('open', 'close'),
+                ('close', 'locked'),
+                ('close', 'open'),
+                ))
         cls._buttons.update({
                 'create_period': {
                     'invisible': ((Eval('state') != 'open')
                         | Eval('periods', [0])),
+                    'depends': ['state'],
                     },
                 'create_period_3': {
                     'invisible': ((Eval('state') != 'open')
                         | Eval('periods', [0])),
+                    'depends': ['state'],
                     },
                 'close': {
                     'invisible': Eval('state') != 'open',
+                    'depends': ['state'],
                     },
                 'reopen': {
                     'invisible': Eval('state') != 'close',
+                    'depends': ['state'],
+                    },
+                'lock': {
+                    'invisible': Eval('state') != 'close',
+                    'depends': ['state'],
                     },
                 })
 
@@ -95,6 +111,13 @@ class FiscalYear(ModelSQL, ModelView):
     def default_company():
         return Transaction().context.get('company')
 
+    def get_icon(self, name):
+        return {
+            'open': 'tryton-open',
+            'close': 'tryton-close',
+            'locked': 'tryton-readonly',
+            }.get(self.state)
+
     @classmethod
     def validate(cls, years):
         super(FiscalYear, cls).validate(years)
@@ -103,7 +126,10 @@ class FiscalYear(ModelSQL, ModelView):
             year.check_post_move_sequence()
 
     def check_dates(self):
-        cursor = Transaction().cursor
+        transaction = Transaction()
+        connection = transaction.connection
+        transaction.database.lock(connection, self._table)
+        cursor = connection.cursor()
         table = self.__table__()
         cursor.execute(*table.select(table.id,
                 where=(((table.start_date <= self.start_date)
@@ -136,7 +162,6 @@ class FiscalYear(ModelSQL, ModelView):
     @classmethod
     def write(cls, *args):
         actions = iter(args)
-        args = []
         for fiscalyears, values in zip(actions, actions):
             if values.get('post_move_sequence'):
                 for fiscalyear in fiscalyears:
@@ -145,12 +170,6 @@ class FiscalYear(ModelSQL, ModelView):
                             values['post_move_sequence']):
                         cls.raise_user_error('change_post_move_sequence', (
                                 fiscalyear.rec_name,))
-            values = values.copy()
-            if 'periods' in values:
-                operator = ['delete', 'remove', 'create', 'write', 'add']
-                values['periods'].sort(lambda x, y:
-                    cmp(operator.index(x[0]), operator.index(y[0])))
-            args.extend((fiscalyears, values))
         super(FiscalYear, cls).write(*args)
 
     @classmethod
@@ -218,23 +237,17 @@ class FiscalYear(ModelSQL, ModelView):
             ], order=[('start_date', 'DESC')], limit=1)
         if not fiscalyears:
             if exception:
-                language = Transaction().language
-                languages = Lang.search([('code', '=', language)])
-                if not languages:
-                    languages = Lang.search([('code', '=', 'en_US')])
-                language, = languages
-                formatted = Lang.strftime(date, language.code, language.date)
-                cls.raise_user_error('no_fiscalyear_date', (formatted,))
+                lang = Lang.get()
+                cls.raise_user_error('no_fiscalyear_date', lang.strftime(date))
             else:
                 return None
         return fiscalyears[0].id
 
-    def _process_account(self, account):
-        '''
-        Process account for a fiscal year closed
-        '''
-        Currency = Pool().get('currency.currency')
-        Deferral = Pool().get('account.account.deferral')
+    def get_deferral(self, account):
+        'Computes deferrals for accounts'
+        pool = Pool()
+        Currency = pool.get('currency.currency')
+        Deferral = pool.get('account.account.deferral')
 
         if account.kind == 'view':
             return
@@ -243,15 +256,17 @@ class FiscalYear(ModelSQL, ModelView):
                 self.raise_user_error('account_balance_not_zero',
                         error_args=(account.rec_name,))
         else:
-            Deferral.create([{
-                        'account': account.id,
-                        'fiscalyear': self.id,
-                        'debit': account.debit,
-                        'credit': account.credit,
-                        }])
+            deferral = Deferral()
+            deferral.account = account
+            deferral.fiscalyear = self
+            deferral.debit = account.debit
+            deferral.credit = account.credit
+            deferral.amount_second_currency = account.amount_second_currency
+            return deferral
 
     @classmethod
     @ModelView.button
+    @Workflow.transition('close')
     def close(cls, fiscalyears):
         '''
         Close a fiscal year
@@ -259,7 +274,15 @@ class FiscalYear(ModelSQL, ModelView):
         pool = Pool()
         Period = pool.get('account.period')
         Account = pool.get('account.account')
+        Deferral = pool.get('account.account.deferral')
+        transaction = Transaction()
+        database = transaction.database
+        connection = transaction.connection
 
+        # Lock period to be sure no new period will be created in between.
+        database.lock(connection, Period._table)
+
+        deferrals = []
         for fiscalyear in fiscalyears:
             if cls.search([
                         ('end_date', '<=', fiscalyear.start_date),
@@ -268,11 +291,6 @@ class FiscalYear(ModelSQL, ModelView):
                         ]):
                 cls.raise_user_error('close_error', (fiscalyear.rec_name,))
 
-            #First close the fiscalyear to be sure
-            #it will not have new period created between.
-            cls.write([fiscalyear], {
-                'state': 'close',
-                })
             periods = Period.search([
                     ('fiscalyear', '=', fiscalyear.id),
                     ])
@@ -283,11 +301,13 @@ class FiscalYear(ModelSQL, ModelView):
                 accounts = Account.search([
                         ('company', '=', fiscalyear.company.id),
                         ])
-            for account in accounts:
-                fiscalyear._process_account(account)
+                deferrals += [_f for _f in (fiscalyear.get_deferral(a)
+                        for a in accounts) if _f]
+        Deferral.save(deferrals)
 
     @classmethod
     @ModelView.button
+    @Workflow.transition('open')
     def reopen(cls, fiscalyears):
         '''
         Re-open a fiscal year
@@ -297,7 +317,7 @@ class FiscalYear(ModelSQL, ModelView):
         for fiscalyear in fiscalyears:
             if cls.search([
                         ('start_date', '>=', fiscalyear.end_date),
-                        ('state', '=', 'close'),
+                        ('state', '!=', 'open'),
                         ('company', '=', fiscalyear.company.id),
                         ]):
                 cls.raise_user_error('reopen_error')
@@ -307,16 +327,16 @@ class FiscalYear(ModelSQL, ModelView):
                 ])
             Deferral.delete(deferrals)
 
-            cls.write([fiscalyear], {
-                'state': 'open',
-                })
-
     @classmethod
-    def search_rec_name(cls, name, clause):
-        return ['OR',
-            ('code',) + tuple(clause[1:]),
-            (cls._rec_name,) + tuple(clause[1:]),
-            ]
+    @ModelView.button
+    @Workflow.transition('locked')
+    def lock(cls, fiscalyears):
+        pool = Pool()
+        Period = pool.get('account.period')
+        periods = Period.search([
+                ('fiscalyear', 'in', [f.id for f in fiscalyears]),
+                ])
+        Period.lock(periods)
 
 
 class BalanceNonDeferralStart(ModelView):
@@ -324,6 +344,8 @@ class BalanceNonDeferralStart(ModelView):
     __name__ = 'account.fiscalyear.balance_non_deferral.start'
     fiscalyear = fields.Many2One('account.fiscalyear', 'Fiscal Year',
         required=True, domain=[('state', '=', 'open')])
+    company = fields.Function(fields.Many2One('company.company', 'Company'),
+        'on_change_with_company')
     journal = fields.Many2One('account.journal', 'Journal', required=True,
         domain=[
             ('type', '=', 'situation'),
@@ -338,16 +360,23 @@ class BalanceNonDeferralStart(ModelView):
         required=True,
         domain=[
             ('kind', '!=', 'view'),
-            ('company', '=', Eval('context', {}).get('company', -1)),
+            ('company', '=', Eval('company', -1)),
             ('deferral', '=', True),
-            ])
+            ],
+        depends=['company'])
     debit_account = fields.Many2One('account.account', 'Debit Account',
         required=True,
         domain=[
             ('kind', '!=', 'view'),
-            ('company', '=', Eval('context', {}).get('company', -1)),
+            ('company', '=', Eval('company', -1)),
             ('deferral', '=', True),
-            ])
+            ],
+        depends=['company'])
+
+    @fields.depends('fiscalyear')
+    def on_change_with_company(self, name=None):
+        if self.fiscalyear:
+            return self.fiscalyear.company.id
 
 
 class BalanceNonDeferral(Wizard):
@@ -356,7 +385,7 @@ class BalanceNonDeferral(Wizard):
     start = StateView('account.fiscalyear.balance_non_deferral.start',
         'account.fiscalyear_balance_non_deferral_start_view_form', [
             Button('Cancel', 'end', 'tryton-cancel'),
-            Button('Ok', 'balance', 'tryton-ok', default=True),
+            Button('OK', 'balance', 'tryton-ok', default=True),
             ])
     balance = StateAction('account.act_move_line_form')
 
@@ -434,25 +463,110 @@ class BalanceNonDeferral(Wizard):
         return action, {}
 
 
-class CloseFiscalYearStart(ModelView):
-    'Close Fiscal Year'
-    __name__ = 'account.fiscalyear.close.start'
-    close_fiscalyear = fields.Many2One('account.fiscalyear',
-            'Fiscal Year to close', required=True,
-            domain=[('state', '!=', 'close')])
+def month_delta(d1, d2):
+    return (d1.year - d2.year) * 12 + d1.month - d2.month
 
 
-class CloseFiscalYear(Wizard):
-    'Close Fiscal Year'
-    __name__ = 'account.fiscalyear.close'
-    start = StateView('account.fiscalyear.close.start',
-        'account.fiscalyear_close_start_view_form', [
-            Button('Cancel', 'end', 'tryton-cancel'),
-            Button('Close', 'close', 'tryton-ok', default=True),
+class RenewFiscalYearStart(ModelView):
+    "Renew Fiscal Year Start"
+    __name__ = 'account.fiscalyear.renew.start'
+    name = fields.Char("Name", required=True)
+    company = fields.Many2One('company.company', "Company", required=True)
+    previous_fiscalyear = fields.Many2One(
+        'account.fiscalyear', "Previous Fiscalyear", required=True,
+        domain=[
+            ('company', '=', Eval('company')),
+            ],
+        depends=['company'],
+        help="Used as reference for fiscalyear configuration.")
+    start_date = fields.Date("Start Date", required=True)
+    end_date = fields.Date("End Date", required=True)
+    reset_sequences = fields.Boolean("Reset Sequences",
+        help="If checked, new sequences will be created.")
+
+    @classmethod
+    def default_company(cls):
+        return Transaction().context.get('company')
+
+    @classmethod
+    def default_previous_fiscalyear(cls):
+        pool = Pool()
+        FiscalYear = pool.get('account.fiscalyear')
+        fiscalyears = FiscalYear.search([
+                ('company', '=', cls.default_company() or -1),
+                ],
+            order=[('end_date', 'DESC')], limit=1)
+        if fiscalyears:
+            fiscalyear, = fiscalyears
+            return fiscalyear.id
+
+    @classmethod
+    def default_reset_sequences(cls):
+        return True
+
+    @fields.depends('previous_fiscalyear')
+    def on_change_previous_fiscalyear(self):
+        if self.previous_fiscalyear:
+            fiscalyear = self.previous_fiscalyear
+            months = month_delta(
+                fiscalyear.end_date, fiscalyear.start_date) + 1
+            self.start_date = fiscalyear.start_date + relativedelta(
+                months=months, day=fiscalyear.start_date.day)
+            self.end_date = fiscalyear.end_date + relativedelta(
+                months=months, day=fiscalyear.end_date.day)
+            self.name = fiscalyear.name.replace(
+                str(fiscalyear.end_date.year),
+                str(self.end_date.year)).replace(
+                str(fiscalyear.start_date.year),
+                str(self.start_date.year))
+
+
+class RenewFiscalYear(Wizard):
+    "Renew Fiscal Year"
+    __name__ = 'account.fiscalyear.renew'
+    start = StateView('account.fiscalyear.renew.start',
+        'account.fiscalyear_renew_start_view_form', [
+            Button("Cancel", 'end', 'tryton-cancel'),
+            Button("Create", 'create_', 'tryton-ok', default=True),
             ])
-    close = StateTransition()
+    create_ = StateAction('account.act_fiscalyear_form')
 
-    def transition_close(self):
-        Fiscalyear = Pool().get('account.fiscalyear')
-        Fiscalyear.close([self.start.close_fiscalyear])
-        return 'end'
+    def fiscalyear_defaults(self):
+        pool = Pool()
+        Sequence = pool.get('ir.sequence')
+        defaults = {
+            'name': self.start.name,
+            'start_date': self.start.start_date,
+            'end_date': self.start.end_date,
+            'periods': [],
+            }
+        previous_sequence = self.start.previous_fiscalyear.post_move_sequence
+        sequence, = Sequence.copy([previous_sequence])
+        if self.start.reset_sequences:
+            sequence.number_next = 1
+        else:
+            sequence.number_next = previous_sequence.number_next
+        sequence.save()
+        defaults['post_move_sequence'] = sequence.id
+        return defaults
+
+    def create_fiscalyear(self):
+        pool = Pool()
+        FiscalYear = pool.get('account.fiscalyear')
+        fiscalyear, = FiscalYear.copy(
+            [self.start.previous_fiscalyear],
+            default=self.fiscalyear_defaults())
+        periods = [p for p in self.start.previous_fiscalyear.periods
+            if p.type == 'standard']
+        months = month_delta(fiscalyear.end_date, fiscalyear.start_date) + 1
+        if len(periods) == months:
+            FiscalYear.create_period([fiscalyear])
+        elif len(periods) == months / 3:
+            FiscalYear.create_period_3([fiscalyear])
+        return fiscalyear
+
+    def do_create_(self, action):
+        fiscalyear = self.create_fiscalyear()
+        action['res_id'] = [fiscalyear.id]
+        action['views'].reverse()
+        return action, {}
