@@ -1208,7 +1208,7 @@ class TaxLine(ModelSQL, ModelView):
                         & (table.code.in_(
                                 [invoice_base_code, credit_note_base_code]))))
 
-    @fields.depends('move_line')
+    @fields.depends('move_line', '_parent_move_line.currency_digits')
     def on_change_with_currency_digits(self, name=None):
         if self.move_line:
             return self.move_line.currency_digits
@@ -1220,7 +1220,10 @@ class TaxLine(ModelSQL, ModelView):
             return self.move_line.account.company.id
 
     def get_rec_name(self, name):
-        return self.tax.rec_name
+        name = super(TaxLine, self).get_rec_name(name)
+        if self.tax:
+            name = self.tax.rec_name
+        return name
 
     @classmethod
     def search_rec_name(cls, name, clause):
